@@ -1,7 +1,7 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2014-2015 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2014-2017 -- leonerd@leonerd.org.uk
 
 package Net::Async::Matrix;
 
@@ -11,7 +11,7 @@ use warnings;
 use base qw( IO::Async::Notifier );
 IO::Async::Notifier->VERSION( '0.63' ); # adopt_future
 
-our $VERSION = '0.18';
+our $VERSION = '0.19';
 $VERSION = eval $VERSION;
 
 use Carp;
@@ -22,6 +22,8 @@ use JSON::MaybeXS qw( encode_json decode_json );
 
 use Data::Dump 'pp';
 use File::stat;
+use List::Util 1.29 qw( pairmap );
+use Scalar::Util qw( blessed );
 use Struct::Dumb;
 use Time::HiRes qw( time );
 use URI;
@@ -30,7 +32,7 @@ struct User => [qw( user_id displayname presence last_active )];
 
 use Net::Async::Matrix::Room;
 
-use constant PATH_PREFIX => "/_matrix/client/api/v1";
+use constant PATH_PREFIX => "/_matrix/client/r0";
 use constant LONGPOLL_TIMEOUT => 30;
 
 # This is only needed for the (undocumented) recaptcha bypass feature
@@ -54,7 +56,7 @@ C<Net::Async::Matrix> - use Matrix with L<IO::Async>
  $loop->add( $matrix );
 
  $matrix->login(
-    user     => '@my-user:home.server',
+    user_id  => '@my-user:home.server',
     password => 'SeKr1t',
  )->get;
 
@@ -158,6 +160,12 @@ in receiveing any events, and wishes to reduce the load on the homeserver.
 Optional. Timeout in seconds for the C</events> longpoll operation. Defaults
 to 30 seconds if not supplied.
 
+=head2 first_sync_limit => NUM
+
+Optional. Number of events per room to fetch on the first C</sync> request on
+startup. Defaults to the server's builtin value if not defined, which is
+likely to be 10.
+
 =cut
 
 sub _init
@@ -210,6 +218,7 @@ sub configure
    my %params = @_;
 
    foreach (qw( server path_prefix ua SSL enable_events longpoll_timeout
+                first_sync_limit
                 on_log on_unknown_event on_presence on_room_new on_room_del on_invite
                 on_room_member on_room_message )) {
       $self->{$_} = delete $params{$_} if exists $params{$_};
@@ -231,6 +240,14 @@ sub log
    $self->{on_log}->( $message ) if $self->{on_log};
 }
 
+sub _maybe_encode
+{
+   my $v = shift;
+   return $v if !ref $v or blessed $v;
+   return $v if ref $v ne "HASH";
+   return encode_json( $v );
+}
+
 sub _uri_for_path
 {
    my $self = shift;
@@ -244,7 +261,9 @@ sub _uri_for_path
    $uri->path( $self->{path_prefix} . $path );
 
    $params{access_token} = $self->{access_token} if defined $self->{access_token};
-   $uri->query_form( %params );
+
+   # Some parameter values can be JSON-encoded objects
+   $uri->query_form( pairmap { $a => _maybe_encode $b } %params );
 
    return $uri;
 }
@@ -371,7 +390,9 @@ sub _do_POST_file
    return $self->adopt_future( $f );
 }
 
-=head2 $matrix->login( %params )->get
+=head2 login
+
+   $matrix->login( %params )->get
 
 Performs the necessary steps required to authenticate with the configured
 Home Server, actually obtain an access token and starting the event stream
@@ -407,8 +428,9 @@ sub login
    if( defined $params{user_id} and defined $params{access_token} ) {
       $self->{$_} = $params{$_} for qw( user_id access_token );
       $self->configure( notifier_name => "uid=$params{user_id}" );
-      $self->start;
-      return Future->done( $self );
+      return ( ( $self->{enable_events} // 1 ) ? $self->start : Future->done )->then( sub {
+         Future->done( $self )
+      });
    }
 
    # Otherwise; try to obtain the login flow information
@@ -442,12 +464,14 @@ sub _login_with_password
       { type => "m.login.password", user => $params{user_id}, password => $params{password} }
    )->then( sub {
       my ( $resp ) = @_;
-      return $self->login( %$resp ) if defined $resp->{access_token};
+      return $self->login( %$resp, %params ) if defined $resp->{access_token};
       return Future->fail( "Expected server to respond with 'access_token'", matrix => );
    });
 }
 
-=head2 $matrix->register( %params )->get
+=head2 register
+
+   $matrix->register( %params )->get
 
 Performs the necessary steps required to create a new account on the
 configured Home Server.
@@ -545,20 +569,49 @@ sub _register_with_recaptcha
    };
 }
 
+=head2 sync
+
+   $matrix->sync( %params )->get
+
+Performs a single C</sync> request on the server, returning the raw results
+directly.
+
+Takes the following named parameters
+
+=over 4
+
+=item since => STRING
+
+Optional. Sync token from the previous request.
+
+=back
+
+=cut
+
+sub sync
+{
+   my $self = shift;
+   my ( %params ) = @_;
+
+   $self->_do_GET_json( "/sync", %params );
+}
+
 sub await_synced
 {
    my $self = shift;
    return $self->{synced_future} //= $self->loop->new_future;
 }
 
-=head2 $f = $matrix->start
+=head2 start
 
-Performs the initial IMSync on the server, and starts the event stream to
+   $f = $matrix->start
+
+Performs the initial sync on the server, and starts the event stream to
 begin receiving events.
 
 While this method does return a C<Future> it is not required that the caller
 keep track of this; the object itself will store it. It will complete when the
-initial IMSync has fininshed, and the event stream has started.
+initial sync has fininshed, and the event stream has started.
 
 If the initial sync has already been requested, this method simply returns the
 future it returned the last time, ensuring that you can await the client
@@ -574,55 +627,31 @@ sub start
 
    return $self->{start_f} ||= do {
       undef $self->{synced_future};
+
       foreach my $room ( values %{ $self->{rooms_by_id} } ) {
          $room->_reset_for_sync;
       }
 
-      my $f = $self->_do_GET_json( "/initialSync", limit => 0 )
-         ->on_fail( sub { undef $self->{start_f} } )
-      ->then( sub {
+      my %first_sync_args;
+
+      $first_sync_args{filter}{room}{timeline}{limit} = $self->{first_sync_limit} 
+         if defined $self->{first_sync_limit};
+
+      $self->sync( %first_sync_args )->then( sub {
          my ( $sync ) = @_;
 
-         foreach ( @{ $sync->{rooms} } ) {
-            my $room_id = $_->{room_id};
-            my $membership = $_->{membership};
+         $self->_incoming_sync( $sync );
 
-            if( $membership eq "join" ) {
-               my $state = $_->{state};
+         $self->start_longpoll( since => $sync->{next_batch} );
 
-               my $room = $self->_get_or_make_room( $room_id );
-
-               ## Placate odd synapse bug wherein some rooms appear twice in
-               ##   initialSync
-               next if $room->await_synced->is_ready;
-
-               $room->_handle_event_initial( $_ ) for @$state;
-
-               $room->await_synced->done;
-               $room->maybe_invoke_event( on_synced_state => );
-            }
-            elsif( $membership eq "invite" ) {
-               $self->maybe_invoke_event( on_invite => $_ );
-            }
-            else {
-               $self->log( "TODO: imsync returned a room in membership state $membership" );
-            }
-         }
-
-         # Now push use presence messages
-         foreach ( @{ $sync->{presence} } ) {
-            $self->_incoming_event( $_ );
-         }
-
-         $self->await_synced->done;
-         $self->start_longpoll( start => $sync->{end} ) if $self->{enable_events} // 1;
-         Future->done;
-      });
-      $self->adopt_future( $f );
+         return $self->await_synced->done;
+      })->on_fail( sub { undef $self->{start_f} });
    };
 }
 
-=head2 $matrix->stop
+=head2 stop
+
+   $matrix->stop
 
 Stops the event stream. After calling this you will need to use C<start> again
 to continue receiving events.
@@ -645,7 +674,7 @@ sub start_longpoll
    my %args = @_;
 
    $self->stop_longpoll;
-   $self->{longpoll_last_token} = $args{start} // "END";
+   $self->{longpoll_last_token} = $args{since};
 
    my $f = $self->{longpoll_f} = repeat {
       my $last_token = $self->{longpoll_last_token};
@@ -654,14 +683,15 @@ sub start_longpoll
          $self->{make_delay}->( $self->{longpoll_timeout} + 5 )
             ->else_fail( "Longpoll timed out" ),
 
-         $self->_do_GET_json( "/events",
-            $last_token ? ( from => $last_token ) : (),
+         $self->sync(
+            since   => $last_token,
             timeout => $self->{longpoll_timeout} * 1000, # msec
          )->then( sub {
-            my ( $data ) = @_;
+            my ( $sync ) = @_;
 
-            $self->_incoming_event( $_ ) foreach @{ $data->{chunk} };
-            $self->{longpoll_last_token} = $data->{end};
+            $self->_incoming_sync( $sync );
+
+            $self->{longpoll_last_token} = $sync->{next_batch};
 
             Future->done();
          }),
@@ -736,7 +766,9 @@ sub _get_or_make_room
       $self->_make_room( $room_id );
 }
 
-=head2 $user = $matrix->myself
+=head2 myself
+
+   $user = $matrix->myself
 
 Returns the user object representing the connected user.
 
@@ -748,7 +780,9 @@ sub myself
    return $self->_get_or_make_user( $self->{user_id} );
 }
 
-=head2 $user = $matrix->user( $user_id )
+=head2 user
+
+   $user = $matrix->user( $user_id )
 
 Returns the user object representing a user of the given ID, if defined, or
 C<undef>.
@@ -762,26 +796,27 @@ sub user
    return $self->{users_by_id}{$user_id};
 }
 
-sub _incoming_event
+sub _incoming_sync
 {
    my $self = shift;
-   my ( $event ) = @_;
+   my ( $sync ) = @_;
 
-   my @type_parts = split m/\./, $event->{type};
-   my @subtype_args;
+   foreach my $category (qw( invite join leave )) {
+      my $rooms = $sync->{rooms}{$category} or next;
+      foreach my $room_id ( keys %$rooms ) {
+         my $roomsync = $rooms->{$room_id};
 
-   while( @type_parts ) {
-      if( my $handler = $self->can( "_handle_event_" . join "_", @type_parts ) ) {
-         $handler->( $self, @subtype_args, $event );
-         return;
+         my $room = $self->_get_or_make_room( $room_id );
+
+         $room->${\"_incoming_sync_$category"}( $roomsync );
       }
-
-      unshift @subtype_args, pop @type_parts;
    }
 
-   $self->maybe_invoke_event(
-      on_unknown_event => $event
-   ) or $self->log( "  incoming event=".pp($event) );
+   foreach my $event ( @{ $sync->{presence}{events} } ) {
+      $self->_handle_event_m_presence( $event );
+   }
+
+   # TODO: account_data
 }
 
 sub _on_self_leave
@@ -794,9 +829,13 @@ sub _on_self_leave
    delete $self->{rooms_by_id}{$room->room_id};
 }
 
-=head2 $name = $matrix->get_displayname->get
+=head2 get_displayname
 
-=head2 $matrix->set_displayname( $name )->get
+=head2 set_displayname
+
+   $name = $matrix->get_displayname->get
+
+   $matrix->set_displayname( $name )->get
 
 Accessor and mutator for the user account's "display name" profile field.
 
@@ -826,9 +865,13 @@ sub set_displayname
    );
 }
 
-=head2 ( $presence, $msg ) = $matrix->get_presence->get
+=head2 get_presence
 
-=head2 $matrix->set_presence( $presence, $msg )->get
+=head2 set_presence
+
+   ( $presence, $msg ) = $matrix->get_presence->get
+
+   $matrix->set_presence( $presence, $msg )->get
 
 Accessor and mutator for the user's current presence state and optional status
 message string.
@@ -899,7 +942,9 @@ sub drop_presence
    );
 }
 
-=head2 ( $room, $room_alias ) = $matrix->create_room( $alias_localpart )->get
+=head2 create_room
+
+   ( $room, $room_alias ) = $matrix->create_room( $alias_localpart )->get
 
 Requests the creation of a new room and associates a new alias with the given
 localpart on the server. The returned C<Future> will return an instance of
@@ -926,7 +971,9 @@ sub create_room
    });
 }
 
-=head2 $room = $matrix->join_room( $room_alias_or_id )->get
+=head2 join_room
+
+   $room = $matrix->join_room( $room_alias_or_id )->get
 
 Requests to join an existing room with the given alias name or plain room ID.
 If this room is already known by the C<$matrix> object, this method simply
@@ -948,8 +995,7 @@ sub join_room
       }
       else {
          my $room = $self->_make_room( $room_id );
-         $room->initial_sync
-            ->then_done( $room );
+         return $room->await_synced->then_done( $room );
       }
    });
 }
@@ -965,9 +1011,13 @@ sub room_list
       });
 }
 
-=head2 $matrix->add_alias( $alias, $room_id )->get
+=head2 add_alias
 
-=head2 $matrix->delete_alias( $alias )->get
+=head2 delete_alias
+
+   $matrix->add_alias( $alias, $room_id )->get
+
+   $matrix->delete_alias( $alias )->get
 
 Performs a directory server request to create the given room alias name, to
 point at the room ID, or to remove it again.
@@ -998,7 +1048,9 @@ sub delete_alias
       ->then_done();
 }
 
-=head2 $content_uri = $matrix->upload( %params )->get
+=head2 upload
+
+   $content_uri = $matrix->upload( %params )->get
 
 Performs a post to the server's media content repository, to upload a new
 piece of content, returning the content URI that points to it.
@@ -1059,6 +1111,36 @@ sub upload
    });
 }
 
+=head2 convert_mxc_url
+
+   $url = $matrix->convert_mxc_url( $mxc )
+
+Given a plain string or L<URI> instance containing a Matrix media URL (in the
+C<mxc:> scheme), returns an C<http:> or C<https:> URL in the form of an L<URI>
+instance pointing at the media repository on the user's local homeserver where
+it can be downloaded from.
+
+=cut
+
+sub convert_mxc_url
+{
+   my $self = shift;
+   my ( $mxc ) = @_;
+
+   ( blessed $mxc and $mxc->isa( "URI" ) ) or
+      $mxc = URI->new( $mxc );
+
+   $mxc->scheme eq "mxc" or
+      croak "Require an mxc:// scheme";
+
+   my $uri = URI->new;
+   $uri->scheme( $self->{SSL} ? "https" : "http" );
+   $uri->authority( $self->{server} );
+   $uri->path( "/_matrix/media/v1/download/" . $mxc->authority . $mxc->path );
+
+   return $uri;
+}
+
 ## Incoming events
 
 sub _handle_event_m_presence
@@ -1067,7 +1149,7 @@ sub _handle_event_m_presence
    my ( $event ) = @_;
    my $content = $event->{content};
 
-   my $user = $self->_get_or_make_user( $content->{user_id} );
+   my $user = $self->_get_or_make_user( $event->{sender} );
 
    my %changes;
    foreach (qw( presence displayname )) {
@@ -1094,77 +1176,6 @@ sub _handle_event_m_presence
    }
 }
 
-sub _handle_event_m_room
-{
-   my $self = shift;
-   my $event = pop;
-   my @type_parts = @_;
-
-   # Room messages for existing rooms
-   my $handler;
-   if( $handler = $self->{rooms_by_id}{$event->{room_id}} ) {
-      # OK
-   }
-   elsif( $event->{state_key} eq $self->{user_id} ) {
-      $handler = $self;
-   }
-   else {
-      $self->log( "TODO: Room event on unknown room ID $event->{room_id} not about myself" );
-      # Ignore it for now
-      return;
-   }
-
-   my @subtype_parts;
-   while( @type_parts ) {
-      if( my $code = $handler->can( "_handle_roomevent_" . join "_", @type_parts, "forward" ) ) {
-         # Avoid a race condition if a new room event arrives before we finish
-         # the room initial sync
-         $handler->await_synced->on_done( sub {
-            $code->( $handler, @subtype_parts, $event );
-         });
-         return;
-      }
-
-      unshift @subtype_parts, pop @type_parts;
-   }
-
-   $self->log( "Unhandled room event " . join( "_", @subtype_parts ) . "\n" .
-      pp( $event ) );
-}
-
-sub _handle_event_m_typing
-{
-   my $self = shift;
-   my $event = pop;
-
-   if( my $room = $self->{rooms_by_id}{$event->{room_id}} ) {
-      $room->_handle_event_m_typing( $event );
-   }
-}
-
-sub _handle_roomevent_member_forward
-{
-   my $self = shift;
-   my $event = pop;
-
-   my $content = $event->{content};
-   my $membership = $content->{membership};
-
-   if( $membership eq "join" ) {
-      my $room = $self->_get_or_make_room( $event->{room_id} );
-      $self->adopt_future(
-         # TODO: "members" isn't enough. We want other config too...
-         $room->initial_sync
-      );
-   }
-   elsif( $membership eq "invite" ) {
-      $self->maybe_invoke_event( on_invite => $event );
-   }
-   else {
-      $self->log( "Unhandled selfroom event member membership=$membership" );
-   }
-}
-
 =head1 USER STRUCTURES
 
 Parameters documented as C<$user> receive a user struct, which supports the
